/*
    This Software (C) Copyright David Goodwin, 2007, 2008.
    
     This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
/*
 * LTS.java
 *
 * Created on 23 July 2007, 23:13
 */

package assignment;

import java.util.*;
import java.lang.*;
import java.io.*;
import logging.*;
import logging.ILogger.Severity;

/** This class represents a Labeled Transition System.
 *
 * @author David
 */
public class LTS {
   
    protected String fileName;
    protected LTSNode startNode;
    protected HashSet<LTSTransition> transitionList;
    
    /** This is used as a parameter to the parallel reach recursive function
     */ 
    private static class parallel_reach_params {
        public LTS ltsA;
        public LTS ltsB;
        public HashSet<String> syncSet;
        
        public HashMap<LTSNode, HashSet<LTSTransition>> nodeMapA;
        public HashMap<LTSNode, HashSet<LTSTransition>> nodeMapB;
        public Hashtable<String,LTSTransition> transitionMapA;
        public Hashtable<String,LTSTransition> transitionMapB;
        
        public Collection<LTSNode> nodesA;
        public Collection<LTSNode> nodesB;
        
        public Collection<LTSTransition> transitionsA;
        public Collection<LTSTransition> transitionsB;
        
        public LTS newLTS;
        
        public HashSet<LTSNode> visitedNodes;
    }
    
    /** Creates a new instance of LTS */
    public LTS() {
        transitionList = new HashSet<LTSTransition>();//new ArrayList<LTSTransition>();
    }
    
    /** Sets the filename in use. This should only really be used by the load function - if you
     * want to save as a different file name use the save(string) function.
     */
    public void setFileName(String filename) {
        fileName = filename;
    }
    
    /** Returns the start node for this LTS
     */
    public LTSNode getStartNode() {
        return startNode;
    }
    
    /** Sets the start node for this LTS
     */
    public void setStartNode(LTSNode start) {
        startNode = start;
    }
    
    /** Returns true if the specified transition exists in this LTS.
     */
    public boolean transitionExists(LTSTransition ltst) {
        return transitionList.contains(ltst);
    }
    
    /** Adds the specified transition to this LTS
     */
    public void addTransition(LTSTransition ltst) {
        if (!transitionList.contains(ltst))
            transitionList.add(ltst);
    }
    
    /** Removes the specified transition from this LTS.
     */
    public void removeTransition(LTSTransition ltst) {
        transitionList.remove(ltst);
    }
      
    /** Removes all transitions from the LTS.
     */
    public void killTransitions() {
        transitionList.clear();
    }
    
    /** Returns a list of all transitions in this LTS.
     */
    public Collection<LTSTransition> getTransitions() {
        return (Collection<LTSTransition>)transitionList.clone();
    }
    
    /** Returns a list of nodes in the LTS.
     * @param includeStart If set to true this function will return the start node
     * even if it is not present in any transitions. If false it will only return
     * the start node if it is present in one or more transitions
     * @return A list of nodes in the LTS
     */
    public Collection<LTSNode> getNodes(boolean includeStart) {
        Collection<LTSTransition> tl = getTransitions();
        ArrayList<LTSNode> output = new ArrayList<LTSNode>();
        
        HashSet<LTSNode> hs = new HashSet<LTSNode>();
        
        Iterator it = tl.iterator();
        
        while (it.hasNext()) {
            LTSTransition lt = (LTSTransition) it.next();
            
            if (lt.getNode1() != null)
                hs.add(lt.getNode1());
            if (lt.getNode2() != null)
                hs.add(lt.getNode2());
        }
        
        ArrayList<LTSNode> nl = new ArrayList<LTSNode>();
        
        nl.addAll(hs);
        
        if ((getStartNode() != null) && includeStart)
            nl.add(getStartNode());
        
        return nl;
    }
    
    /** This function is for compatibility with older code. It should not be used.
     * It just calls getNodes with includeStart = true.
     * @return A list of all nodes in the LTS (including the start node even if
     * it isnt a member of a transition.
     */
    public Collection<LTSNode> getNodes() {
        return getNodes(true);
    }
    
    /** Loads an LTS from a text file.
     */
    public static LTS load(String filename) throws IOException {
        globals.log.writeLine("LTS: Attempting to load file \"" + filename +"\"");
        
        LTS theLTS = new LTS();
        theLTS.setFileName(filename);
        
        FileInputStream fin;		

        fin = new FileInputStream (filename);
        BufferedReader breader = new BufferedReader(new InputStreamReader(fin));            
        
        String startNodeName = "";
        
        String curlin = breader.readLine();
        
        if (curlin == "") {
            globals.log.writeLine(Severity.Error, "Failed to load file: Invalid format (first line is blank)");
            return null;
        }
        
        while (curlin != null) {
            if (curlin.startsWith("#")) {
                // Its a comment, do nothing
            }
            else {
                if (startNodeName == "") {
                    startNodeName = curlin;
                    
                    // Check the input string is valid
                    startNodeName = checkIString(startNodeName);
                    if (startNodeName == null) return null;
                    
                    theLTS.setStartNode(new LTSNode(startNodeName));
                }
                else {
                    String node1;
                    String node2;
                    String name;

                    String[] stringList = curlin.split("/");
                    
                    if (stringList.length == 2 || stringList.length > 3) {
                        // This file is invalid or something else entirely. We
                        // cant do anything much with it.
                        globals.log.writeLine(Severity.Error, "Failed to load file: Invalid format (invalid entry)");
                        return null;
                    }

                    node1 = stringList[0];
                                       
                    // Check the input string is valid
                    node1 = checkIString(node1);
                    if (node1 == null) return null;
                    
                    if (stringList.length > 1) {
                        name = stringList[1];
                        node2 = stringList[2];

                        // Check the input string is valid
                        node2 = checkIString(node2);
                        if (node2 == null) return null;

                        theLTS.addTransition(new LTSTransition(new LTSNode(node1),new LTSNode(node2),name));
                    }
                    else {                        
                        theLTS.addTransition(new LTSTransition(new LTSNode(node1),null,""));
                    }
                }
            }
            
            curlin = breader.readLine();
        }

        fin.close();		

        return theLTS;
    }
    
    /** Creates a new LTS based on given data
     * 
     * @param content The LTS content. This is of the same format as the text
     * files. You could essentially read the text file into an array list and
     * pass it straight into this function.
     * @return A new LTS based on the given content.
     */
    public static LTS create(ArrayList<String> content) {
        globals.log.writeLine("LTS: Attempting to create new LTS from entered data");
        
        LTS theLTS = new LTS();
        //theLTS.setFileName("");
        
        String startNodeName = "";
        
        String curlin = content.get(0);
        int i = 0;
        while (curlin != null) {
            if (curlin.startsWith("#")) {
                // Its a comment, do nothing
            }
            else {
                if (startNodeName == "") {
                    startNodeName = curlin;
                    
                    // Check the input string is valid
                    startNodeName = checkIString(startNodeName);
                    if (startNodeName == null) return null;
                    
                    theLTS.setStartNode(new LTSNode(startNodeName));
                }
                else {
                    String node1;
                    String node2;
                    String name;

                    String[] stringList = curlin.split("/");
                    
                    if (stringList.length == 2 || stringList.length > 3) {
                        // This file is invalid or something else entirely. We
                        // cant do anything much with it.
                        globals.log.writeLine(Severity.Error, "Failed to load file: Invalid format (invalid entry)");
                        return null;
                    }

                    node1 = stringList[0];
                                       
                    // Check the input string is valid
                    node1 = checkIString(node1);
                    if (node1 == null) return null;
                    
                    if (stringList.length > 1) {
                        name = stringList[1];
                        node2 = stringList[2];

                        // Check the input string is valid
                        node2 = checkIString(node2);
                        if (node2 == null) return null;

                        theLTS.addTransition(new LTSTransition(new LTSNode(node1),new LTSNode(node2),name));
                    }
                    else {                        
                        theLTS.addTransition(new LTSTransition(new LTSNode(node1),null,""));
                    }
                }
            }
            
            if (i < content.size()) {
                curlin = content.get(i);
                i++;
            }
            else
                curlin = null;
        }

       return theLTS;
    }
    
    /** Saves the LTS to the current file ("Save")
     */
    public void save() throws IOException {
        save(getFileName());
    }
    
    /** Saves the LTS to the specified file ("Save As")
     */
    public void save(String filename) throws IOException {
        setFileName(filename);  // Just in case the file name is different.
        FileOutputStream fout;		


        fout = new FileOutputStream (getFileName());
        PrintStream ps = new PrintStream(fout);
        
        // Print the start node:
        ps.println (getStartNode().getName());
        
        // Print the transitions:
        for (LTSTransition t:transitionList) {
            ps.println(t.toString());
        }

        fout.close();		

    }
    
    /** Compares the LTS with another.
     *@return true if the two LTS are equal.
     */
    public boolean equals(LTS theLTS){
        // Ensure they have the same start nodes
        String lSN = theLTS.getStartNode().getName();
        String lSNN = "";
        if (getStartNode() != null)
            lSNN = getStartNode().getName();
        if (!lSN.equals(lSNN))
            return false;
        
        // If one has more transitions than the other they cant be equal.
        if (theLTS.getTransitions().size() != getTransitions().size())
            return false;
        
        // Ensure they have the same transitions
        for (LTSTransition t:transitionList){
            boolean te = theLTS.transitionExists(t);
            if (!te)
                return false;
        }
        return true;
    }
    
    /** Composes two LTS and returns the result. Clones lts_a and the composes it with lts_b returning the result.
     * @return The composition of lts_a and lts_b using the given Sync Set.
     */
    public static LTS compose(LTS lts_a, LTS lts_b, HashSet<String> syncSet) {
        
        // We cant have a null SyncSet now can we...
        if (syncSet == null) syncSet = new HashSet<String>();
       
        LTS theNewLTS = new LTS();
        
        // First up, get the nodes from each LTS
        Collection<LTSNode> nA = lts_a.getNodes(false);
        Collection<LTSNode> nB = lts_b.getNodes(false);
        
        Collection<LTSTransition> ts = lts_a.getTransitions();
        
        // In the interest of performance, we must compute a hashtable of transitions.
        Hashtable<String,LTSTransition> htA = new Hashtable<String,LTSTransition>();
        Hashtable<String,LTSTransition> htB = new Hashtable<String,LTSTransition>();
        for (LTSTransition t:ts) {
            htA.put(t.getName(),t);
        }
        for (LTSTransition t:lts_b.getTransitions()) {
            htB.put(t.getName(),t);
        }
        

        
        
        // Now compute the transitions for the transitions from the first LTS and the nodes from the second LTS
        for (LTSTransition t: ts) {
            // Ok, we must check to see if this transition is in both LTS and the sync set.
            if (syncSet.contains(t.getName()) && htB.containsKey(t.getName())) {
                // It is in all 3 sets.
                LTSTransition b = htB.get(t.getName());
                
                LTSTransition lt = new LTSTransition();
                lt.setName(t.getName());
                lt.setNode1(new LTSNode("(" + t.getNode1().getName() + "," + b.getNode1().getName() + ")"));
                lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + b.getNode2().getName() + ")"));
                theNewLTS.addTransition(lt);
                continue;
            }
            else {
                // treat it normally.
                if (t.getNode2() != null)
                    for (LTSNode n : nB) {
                        LTSTransition lt = new LTSTransition();
                        lt.setName(t.getName());
                        lt.setNode1(new LTSNode("(" + t.getNode1().getName() + "," + n.getName() + ")"));
                        lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + n.getName() + ")"));
                        theNewLTS.addTransition(lt);
                    }
            }
        }
        
        ts = lts_b.getTransitions();
       
        // Now compute the transitions for the transitions from the second LTS and the nodes from the first LTS
        for (LTSTransition t: ts) {
            if (syncSet.contains(t.getName()) && htA.containsKey(t.getName())) {
                // The node is synchronized and has already been added. Nothing to do here...
                continue;
            }
            else {
                // some transitions have null as their second node.
                // These really just represent a single node - not a transition.
                if (t.getNode2() != null)
                    for (LTSNode n : nA) {
                        LTSTransition lt = new LTSTransition();
                        lt.setName(t.getName());
                        lt.setNode1(new LTSNode("(" + n.getName() + "," + t.getNode1().getName() + ")"));
                        lt.setNode2(new LTSNode("(" + n.getName() + "," + t.getNode2().getName() + ")"));
                        theNewLTS.addTransition(lt);
                    }
            }
        }
        
        // Lastly, set the start node
        theNewLTS.setStartNode(new LTSNode("(" + lts_a.getStartNode().getName() + "," + lts_b.getStartNode().getName() + ")"));
        
        
        return theNewLTS;
    }
    /** This converts the String syncSet to a HashSet and passes it to the normal compose function.
     * @return The composition of lts_a and lts_b using the given Sync Set.
     */
    public static LTS composeA(LTS lts_a, LTS lts_b, String syncSet) {
        
        if (syncSet == "")
            return compose(lts_a,lts_b,null);
        HashSet<String> hs = new HashSet<String>();
        
        // First we must scan through and strip out any bracketed items:
        if (syncSet.contains("(") && syncSet.contains(")")) {
            // There are brackets. We must strip these items out first.
            String parts = "";
            boolean inBrackets = false;
            int startPosition = 0;
            for (int i = 0; i < syncSet.length(); i++) {
                if (syncSet.charAt(i) == '(')  {
                    inBrackets = true;
                    startPosition = i-1;
                    
                }
                if (inBrackets)                         parts += syncSet.charAt(i);
                if (syncSet.charAt(i) == ')')  {
                    inBrackets = false;
                    hs.add(parts);
                    parts = "";
                    String tmp = syncSet;
                    syncSet = tmp.substring(0,startPosition) + tmp.substring(i+1);
                }
            }
        }
        
        // Convert the string into a HashSet<String>:
        String[] result = syncSet.split(",");
        
        
        for (String s:result) {
            hs.add(s);
        }
        
        return compose(lts_a,lts_b,hs);
    }
    
    /** Performs Parallel Reach. Does Parallel Composition while excluding unreachable nodes
     */
    public static LTS Parallel_Reach(LTS lts_a, LTS lts_b, HashSet<String> syncSet) {
        
        // Cant have a null syncSet
        if (syncSet == null) syncSet = new HashSet<String>();
        
        
        // Prepare the parameters:
        parallel_reach_params params = new parallel_reach_params();
        params.syncSet = syncSet;
        params.nodeMapA = buildNodeMap(lts_a);
        params.nodeMapB = buildNodeMap(lts_b);
        params.ltsA = lts_a;
        params.ltsB = lts_b;
        params.transitionMapA = new Hashtable<String,LTSTransition>();
        params.transitionMapB = new Hashtable<String,LTSTransition>();
        params.newLTS = new LTS();
        params.nodesA = lts_a.getNodes(false);
        params.nodesB = lts_b.getNodes(false);
        params.transitionsA = lts_a.getTransitions();
        params.transitionsB = lts_b.getTransitions();
        params.visitedNodes = new HashSet<LTSNode>();
       
        for (LTSTransition t: params.transitionsA) {
            params.transitionMapA.put(t.getName(),t);
        }
        for (LTSTransition t: params.transitionsB) {
            params.transitionMapB.put(t.getName(),t);
        }
        
        
        // And now we do the composing stuff.
        doReachStuff(lts_a.getStartNode(), lts_b.getStartNode(), params);
        
        return params.newLTS;
    }
    
    /** Does Parallel Reach stuff
     */
    private static void doReachStuff(LTSNode posA, LTSNode posB, parallel_reach_params params) {
        
        // We dont want to get stuck in a loop! make sure we havent been here before:
        LTSNode tmp = new LTSNode("(" + posA.getName() + "," + posB.getName() + ")");
        if (params.visitedNodes.contains(tmp)) return;
        else params.visitedNodes.add(tmp);
        
        // Transitions we can compose
        HashSet<LTSTransition> tA = params.nodeMapA.get(posA);
        HashSet<LTSTransition> tB = params.nodeMapB.get(posB);
        
        // Nodes we can compose with
        HashSet<LTSNode> pA = new HashSet<LTSNode>();
        HashSet<LTSNode> pB = new HashSet<LTSNode>();
        
        pB.add(posB);
        if (tB != null)
        for(LTSTransition t: tB) {
            if (t.getNode2() != null)
                pB.add(t.getNode2());
        }

        pA.add(posA);
        if (tA != null)
        for(LTSTransition t: tA) {
            if (t.getNode2() != null)
                pA.add(t.getNode2());
        }
        
        Hashtable<String,LTSTransition> htA = new Hashtable<String,LTSTransition>();
        Hashtable<String,LTSTransition> htB = new Hashtable<String,LTSTransition>();
        if (tB != null)
        for (LTSTransition t:tB) {
            htB.put(t.getName(),t);
        }
        if (tA != null)
        for (LTSTransition t:tA) {
            htA.put(t.getName(),t);
        }
        
        // Do the first lot of transitions and nodes
        if (tA != null) {
             // Now compute the transitions for the transitions from the first LTS and the nodes from the second LTS
            for (LTSTransition t: tA) {
                if (params.syncSet.contains(t.getName()) && htB.containsKey(t.getName())) {
                    LTSTransition b = htB.get(t.getName());

                    LTSTransition lt = new LTSTransition();
                    lt.setName(t.getName());
           
                    LTSNode nn = new LTSNode("(" + t.getNode1().getName() + "," + htB.get(t.getName()).getNode1().getName() + ")");
                    if (params.newLTS.getStartNode() == null) params.newLTS.setStartNode(nn); 
                    lt.setNode1(nn);
                    
                    lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + htB.get(t.getName()).getNode2().getName() + ")"));
                    params.newLTS.addTransition(lt);
                    
                    doReachStuff(t.getNode2(), htB.get(t.getName()).getNode2(), params);
                    
                    continue;
                }
                else {
                    if (t.getNode2() != null)
                        for (LTSNode n:pB) {
                            LTSTransition lt = new LTSTransition();
                            lt.setName(t.getName());
                            LTSNode nn = new LTSNode("(" + t.getNode1().getName() + "," + n.getName() + ")");
                            if (params.newLTS.getStartNode() == null) params.newLTS.setStartNode(nn); 
                            lt.setNode1(nn);
                            lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + n.getName() + ")"));
                            params.newLTS.addTransition(lt);

                           // Go go recursive stuff! if this were scheme we would need more brackets()
                           doReachStuff(t.getNode2(), n, params);
                        }
                }
            }
        }
        
        // Cant compose with nothing...
        if (tB == null) return;
       
        // And the second lot of transitions and nodes.      
        for (LTSTransition t: tB) {
             if (params.syncSet.contains(t.getName()) && htA.containsKey(t.getName())) {
                 // we have already synced.
                 continue;
             }
             else {
                if (t.getNode2() != null)
                    for(LTSNode n:pA) {
                        LTSTransition lt = new LTSTransition();
                        lt.setName(t.getName());
                        lt.setNode1(new LTSNode("(" + n.getName() + "," + t.getNode1().getName() + ")"));
                        lt.setNode2(new LTSNode("(" + n.getName() + "," + t.getNode2().getName() + ")"));
                        params.newLTS.addTransition(lt);

                        // Go go recursive stuff
                        doReachStuff(n, t.getNode2(), params);
                    }
             }
            }
    }
    
    /** Builds a HashMap using the first node as the key and a HashSet of transitions with
     * that node as the start node for the value.
     */
    private static HashMap<LTSNode, HashSet<LTSTransition>> buildNodeMap(LTS theLTS) {
        HashMap<LTSNode, HashSet<LTSTransition>> tmp = new HashMap<LTSNode, HashSet<LTSTransition>>();
        
        for (LTSTransition t:theLTS.getTransitions()) {
            
            if (!tmp.containsKey(t.getNode1())){
                tmp.put(t.getNode1(), new HashSet<LTSTransition>());
            }
            
            tmp.get(t.getNode1()).add(t);
        }
        
        return tmp;
    }
    
    /** Convenience overload function: does parallel reach using a string
     * as the sync set instead of a hash set.
     */
    public static LTS Parallel_ReachA(LTS lts_a, LTS lts_b, String syncSet) {
        
        if (syncSet == "")
            return compose(lts_a,lts_b,null);
        HashSet<String> hs = new HashSet<String>();
        
        // First we must scan through and strip out any bracketed items:
        if (syncSet.contains("(") && syncSet.contains(")")) {
            // There are brackets. We must strip these items out first.
            String parts = "";
            boolean inBrackets = false;
            int startPosition = 0;
            for (int i = 0; i < syncSet.length(); i++) {
                if (syncSet.charAt(i) == '(')  {
                    inBrackets = true;
                    startPosition = i-1;
                    
                }
                if (inBrackets)                         parts += syncSet.charAt(i);
                if (syncSet.charAt(i) == ')')  {
                    inBrackets = false;
                    hs.add(parts);
                    parts = "";
                    String tmp = syncSet;
                    syncSet = tmp.substring(0,startPosition) + tmp.substring(i+1);
                }
            }
        }
        
        // Convert the string into a HashSet<String>:
        String[] result = syncSet.split(",");
        
        
        for (String s:result) {
            hs.add(s);
        }
        
        return Parallel_Reach(lts_a,lts_b,hs);
    }
    
    /** Returns the name of the file this transition was loaded from or saved to.
     */
    public String getFileName() {
        return fileName;
    }
   
    /** This is a function for debugging purposes
     */
    public static void printLTS(LTS theLTS) {
        globals.log.writeLine("======Dumping LTS======");
        globals.log.writeLine("StartNode: " + theLTS.getStartNode().toString());
        globals.log.writeLine("Transition Count: " + theLTS.getTransitions().size());
        globals.log.writeLine("Transitions: --");
        
        Collection<LTSTransition> ltst = theLTS.getTransitions();
        
        for (LTSTransition lt : ltst) {
            globals.log.writeLine("   " + lt.toString());
        }
        globals.log.writeLine("===LTS Dump Complete===");
        
    }
    
    /** This class holds data for the LTS Coloring algorithms next sets.
     */
    private static class NextThing {
        String transitionName;
        int color;
        
        public boolean equals(Object o) {
              if (o == null) return false;
              if (!(o instanceof NextThing)) return false;
              NextThing n = (NextThing) o;
              if (n.transitionName.equals (transitionName) && n.color == color) return true;
             return false;
        }

        public int hashCode() {
            return (transitionName + color).hashCode();
        }
        public String toString() {
            return "(" + transitionName + ", " + color + ")";
        }
    }
    
    /** This algorithm colors an LTS using the algorithm given in assignment 3
     */
    public static HashMap<LTSNode, Integer> colorLTS(LTS theLTS) {
        globals.log.writeLine("Coloring LTS...");
        // Get the nodes and transitions
        HashSet<LTSNode> N = new HashSet<LTSNode>(theLTS.getNodes());
        HashSet<LTSTransition> T = new HashSet<LTSTransition>(theLTS.getTransitions());
        int highestColor = 1;
        
        // Construct some stuff, set all node colors to 1
        HashMap<LTSNode, Integer> nodeMap = new HashMap<LTSNode, Integer>();
        LTSNode[] nodes = new LTSNode[N.size()];
        int[] color = new int[N.size()];
        int i = 0;
        for (LTSNode n:N) {
            nodes[i] = n;
            color[i] = 0;
            nodeMap.put(n,i);
            i++;
        }
        
        // Group all transitions by their start node:
        HashMap<LTSNode, HashSet<LTSTransition>> nodeTransitionMap = new HashMap<LTSNode, HashSet<LTSTransition>>();
        for (LTSTransition t:T) {
            if (!nodeTransitionMap.containsKey(t.getNode1())) {
                nodeTransitionMap.put(t.getNode1(), new HashSet<LTSTransition>());
            }
            nodeTransitionMap.get(t.getNode1()).add(t);
        }
        
        HashMap<LTSNode,HashSet<NextThing>> Next = new HashMap<LTSNode,HashSet<NextThing>>();
        
        // And now for the coloring algorithm
        it: while (true) {
            // Recalculate the next set:
            // 1: For each node n, compute Next(n) = {(a, Col(m)|n-a->m}
            //Next.clear();
            for (LTSNode n:N) {
                // 1: Get all transitions going out of n
                HashSet<LTSTransition> transitions = nodeTransitionMap.get(n);

                // 2: Create the next set for the current node:
                HashSet<NextThing> nextSet = new HashSet<NextThing>();
                
                // 3: For each transition, store the label and the color of m
                if (transitions != null) // the end nodes dont have any transitions coming out of them
                    for (LTSTransition t:transitions) {
                        if (t.getNode2() != null){
                            NextThing nt = new NextThing();
                            nt.color = color[nodeMap.get(t.getNode2())];
                            nt.transitionName = t.getName();
                            nextSet.add(nt);
                        }
                    }
                else globals.log.writeLine(Severity.Debug, "No transitions for node n=" + n.toString());
                // 4: Store the next set for the current node:
                Next.put(n,nextSet);
            }
            
            /* Some debugging stuff: Dump the next set to the console */
            StringBuilder sb = new StringBuilder();
            sb.append("Next Sets for this iteration:\n");
            for (LTSNode n:N) {
                //System.out.println(Next.get(n));
                sb.append(Next.get(n) + "\n");
            }
            //System.out.println("Dump complete.\n-----------------------");
            sb.append("Dump complete.\n-----------------------\n");
            globals.log.writeLine(Severity.Debug, sb.toString());
            
            
            // for all values of n & m in the set of nodes...
            boolean stop = true;
            theLoops: for (LTSNode n:N) {
                for (LTSNode m:N) {
                    boolean condA = Next.get(n).equals(Next.get(m));
                    boolean condB = color[nodeMap.get(n)] == color[nodeMap.get(m)];
                    if ((condA != condB)) {
                        stop=false;
                        break theLoops;
                    }
                }
            }
            if (stop) break it;
            
            /* Recolor the graph such that, for all values of n & m in N,
             * Color(n) = Color(m) if and only if Next(n) = Next(m)
             */
            for (LTSNode n:N) {
                for (LTSNode m:N) {
                    
                    HashSet<NextThing> nsA = Next.get(n);
                    HashSet<NextThing> nsB = Next.get(m);
                    
                    boolean conditionA = (nsA.equals(nsB));
                    boolean conditionB = (color[nodeMap.get(n)] == color[nodeMap.get(m)]);
                    
                    if (conditionA == conditionB) continue;
                    else if (conditionA) color[nodeMap.get(n)] = color[nodeMap.get(m)];
                    else {
                        color[nodeMap.get(n)] = highestColor++;
                        color[nodeMap.get(m)] = highestColor++;
                    }
                }
            }
        }
        
        // Now prepare and return the result
        HashMap<LTSNode, Integer> result = new HashMap<LTSNode, Integer>();
        for (LTSNode n:N)
            result.put(n, color[nodeMap.get(n)]);
        globals.log.writeLine("Coloring complete.");
        return result;
    }
    
    /** Bisimulation of two LTS
     * @param lts_a The first LTS
     * @param lts_b The second LTS
     * @return true if the two LTS are equivalent
     */
    public static boolean bisimEquals(LTS lts_a, LTS lts_b) {
        
        globals.log.writeLine("Coloring LTS...");
        // Get the nodes and transitions
        HashSet<LTSNode> N = new HashSet<LTSNode>(lts_a.getNodes());
        N.addAll(lts_b.getNodes());
        HashSet<LTSTransition> T = new HashSet<LTSTransition>(lts_a.getTransitions());
        T.addAll(lts_b.getTransitions());
        int highestColor = 1;
        
        // Construct some stuff, set all node colors to 1
        HashMap<LTSNode, Integer> nodeMap = new HashMap<LTSNode, Integer>();
        LTSNode[] nodes = new LTSNode[N.size()];
        int[] color = new int[N.size()];
        int i = 0;
        for (LTSNode n:N) {
            nodes[i] = n;
            color[i] = 0;
            nodeMap.put(n,i);
            i++;
        }
        
        // Group all transitions by their start node:
        HashMap<LTSNode, HashSet<LTSTransition>> nodeTransitionMap = new HashMap<LTSNode, HashSet<LTSTransition>>();
        for (LTSTransition t:T) {
            if (!nodeTransitionMap.containsKey(t.getNode1())) {
                nodeTransitionMap.put(t.getNode1(), new HashSet<LTSTransition>());
            }
            nodeTransitionMap.get(t.getNode1()).add(t);
        }
        
        HashMap<LTSNode,HashSet<NextThing>> Next = new HashMap<LTSNode,HashSet<NextThing>>();
        
        // And now for the coloring algorithm
        it: while (true) {
            // Recalculate the next set:
            // 1: For each node n, compute Next(n) = {(a, Col(m)|n-a->m}
            //Next.clear();
            for (LTSNode n:N) {
                // 1: Get all transitions going out of n
                HashSet<LTSTransition> transitions = nodeTransitionMap.get(n);

                // 2: Create the next set for the current node:
                HashSet<NextThing> nextSet = new HashSet<NextThing>();
                
                // 3: For each transition, store the label and the color of m
                if (transitions != null) // the end nodes dont have any transitions coming out of them
                    for (LTSTransition t:transitions) {
                        if (t.getNode2() != null){
                            NextThing nt = new NextThing();
                            nt.color = color[nodeMap.get(t.getNode2())];
                            nt.transitionName = t.getName();
                            nextSet.add(nt);
                        }
                    }
                else globals.log.writeLine(Severity.Debug, "No transitions for node n=" + n.toString());
                // 4: Store the next set for the current node:
                Next.put(n,nextSet);
            }
            
            /* Some debugging stuff: Dump the next set to the console */
            /*System.out.println("Next Sets for this iteration:");
            for (LTSNode n:N) {
                System.out.println(Next.get(n));
            }
            System.out.println("Dump complete.\n-----------------------");*/

            /* Some debugging stuff: Dump the next set to the console */
            StringBuilder sb = new StringBuilder();
            sb.append("Next Sets for this iteration:\n");
            for (LTSNode n:N) {
                sb.append(Next.get(n) + "\n");
            }
            sb.append("Dump complete.\n-----------------------\n");
            globals.log.writeLine(Severity.Debug, sb.toString());            
            
            
            
            // for all values of n & m in the set of nodes...
            boolean stop = true;
            theLoops: for (LTSNode n:N) {
                for (LTSNode m:N) {
                    boolean condA = Next.get(n).equals(Next.get(m));
                    boolean condB = color[nodeMap.get(n)] == color[nodeMap.get(m)];
                    if ((condA != condB)) {
                        stop=false;
                        break theLoops;
                    }
                }
            }
            if (stop) break it;
            
            /* Recolor the graph such that, for all values of n & m in N,
             * Color(n) = Color(m) if and only if Next(n) = Next(m)
             */
            for (LTSNode n:N) {
                for (LTSNode m:N) {
                    
                    HashSet<NextThing> nsA = Next.get(n);
                    HashSet<NextThing> nsB = Next.get(m);
                    
                    boolean conditionA = (nsA.equals(nsB));
                    boolean conditionB = (color[nodeMap.get(n)] == color[nodeMap.get(m)]);
                    
                    if (conditionA == conditionB) continue;
                    else if (conditionA) color[nodeMap.get(n)] = color[nodeMap.get(m)];
                    else {
                        color[nodeMap.get(n)] = highestColor++;
                        color[nodeMap.get(m)] = highestColor++;
                    }
                }
            }
        }
        
        return (color[nodeMap.get(lts_a.getStartNode())] == color[nodeMap.get(lts_b.getStartNode())]);
    }
    
    /** What does this do? not entirely sure. Maybe it simplifies as coloured, 
     * what ever that means...
     */
    public static LTS simplify(LTS theLTS) {
        // Colour the node graph first:
        HashMap<LTSNode, Integer> colors = LTS.colorLTS(theLTS);
        
        // And then what?
        // Take a coloured graph and produce a new graph by identifying all 
        // coloured edges
        
        return theLTS;
    }
    
    /** Checks an input string is valid. If the input string is not valid
     * then it returns null, otherwise it unquotes it and returns it.
     * @param input The input String
     * @return Either null or the unquoted input string.
     */
    private static String checkIString(String input) {
        // Does the string begin and end with quotes?
        if (!input.startsWith("\"") && !input.endsWith("\"")) {
            // no, it does not.
            // If we encounter any of the following characters
            // the string is invalid:
            //  [space] - | =
            if (input.contains(" ")) {
                globals.log.writeLine(Severity.Error, "Not allowed an unquoted space");
                return null;
            }
            if (input.contains("-")) {
                globals.log.writeLine(Severity.Error, "Not allowed an unquoted \"-\"");
                return null;
            }
            if (input.contains("|")) {
                globals.log.writeLine(Severity.Error, "Not allowed an unquoted \"|\"");
                return null;
            }                        
            if (input.contains("=")) {
                globals.log.writeLine(Severity.Error, "Not allowed an unquoted \"=\"");
                return null;
            }
        }
        // If the string is properly quoted...
        if (input.startsWith("\"") && input.endsWith("\"")) {
            // Then remove the quotes.
            input = input.substring(1, input.length());
        }   
        return input;
    }
}























