/*
    This Software (C) Copyright David Goodwin, 2007, 2008.
    
     This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
        
package logging;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;

/**
 * This class implements the ILogger interface to provide.
 * logging abilities. This logger does NOT write debug messages.
 * 
 * Class: Logger
 * Package: assignment
 * Purpose: To provide logging capabilities by implementing the ILogger
 * interface.
 * @author David Goodwin
 */
public class Logger implements ILogger {
    
    /** The Log File Output Stream.
     */
    protected FileOutputStream mLogFile;
    
    /** The print stream for the log file.
     */
    protected PrintStream mPS;
    
    /** Opens the log file for input.
     * @param filename The output file for logging.
     */
    public void open(final String filename) throws IOException {
        if (mLogFile == null) {
            mLogFile = new FileOutputStream(filename);
            mPS = new PrintStream(mLogFile);
        } else {
            throw new IOException("Log File already open");
        }
    }
    
    /** Closes any open log files.
     */
    public void close() throws IOException {
        if (mLogFile != null) {
            // Flush the cache, close the file and trash the FOS
            mPS.flush();
            mPS.close();
            mPS = null;
            mLogFile.flush();
            mLogFile.close();
            mLogFile = null;
        } else {
            throw new IOException("Log File not open");
        }
            
    }
    
    /** Writes a message to the log file using the Notice severity.
     * @param message The message to write to the log file.
     */
    public void writeLine(final String message) {
        try {
            writeLineE(Severity.Notice, message);
        } catch(final IOException e) {
            System.err.println(e.getMessage());
        }
    }
    
    /** Writes a line to the log file using writeLineE catching any exceptions.
     * 
     * @param sv Severity Level
     * @param message The log message
     */
    public void writeLine(final Severity sv, final String message) {
        try {
            writeLineE(sv, message);
        } catch(final IOException e) {
            System.err.println(e.getMessage());
        }
    }
    
    /** Writes an entry to the log file using the specified severity.
     *  This implementation ignores all Debug severity messages.
     * @param sv Log entry severity
     * @param message Log message
     */
    public void writeLineE(final Severity sv, final String message) throws IOException {
        if (mLogFile == null) {
            System.err.println("ERROR: No log file open. Message is: " + message);
            throw new IOException("No log file open");
        } else {    
            String msg;
            switch (sv) {
                case Notice:
                    msg = svToString(sv) + message;
                    System.out.println(msg);
                    mPS.println(msg);
                    break;
                case Debug:
                    // This logger ignores debugging
                    break;
                case Warning:
                    msg = svToString(sv) + message;
                    System.err.println(msg);
                    mPS.println(msg);
                    break;
                case Error:
                    msg = svToString(sv) + message;
                    System.err.println(msg);
                    mPS.println(msg);
                    break;
                case Failure:
                    msg = svToString(sv) + message;
                    System.err.println(msg);
                    mPS.println(msg);
                    break;
                default:
                    msg = message;
                    System.err.println(msg);
                    mPS.println(msg);
                    break;
            }

        }
    }
    
    /** Returns the String representation of the specified Severity Level. This
     * is the String representation used by the log output.
     * @param sv The severity level
     * @return The string representation of the given Severity Level.
     */
    public String svToString(final Severity sv) {
        switch (sv) {
            case Debug: 
                return "DEBUG: ";
            case Error: 
                return "ERROR: ";
            case Failure: 
                return "FAILURE: ";
            case Notice: 
                return "";
            case Warning: 
                return "WARNING: ";
            default:
                return null;
        }
    }
}
